/***********************************************************************************************************************
 *
 * blueMarine Metadata - open source media workflow
 * Copyright (C) 2007-2011 by Tidalwave s.a.s. (http://www.tidalwave.it)
 *
 ***********************************************************************************************************************
 *
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software distributed under the License is distributed on
 * an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.  See the License for the
 * specific language governing permissions and limitations under the License.
 *
 ***********************************************************************************************************************
 *
 * WWW: http://bluemarine.tidalwave.it
 * SCM: https://kenai.com/hg/bluemarine~metadata-src
 *
 **********************************************************************************************************************/
package it.tidalwave.metadata.iptc.vocabularies.impl;

import java.util.HashMap;
import java.util.Map;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.ObjectOutputStream;
import java.io.InputStream;
import java.io.IOException;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.xpath.XPath;
import javax.xml.xpath.XPathConstants;
import javax.xml.xpath.XPathExpression;
import javax.xml.xpath.XPathExpressionException;
import javax.xml.xpath.XPathFactory;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;
import it.tidalwave.metadata.iptc.vocabularies.IPTCVocabulary.Entry;
import it.tidalwave.metadata.iptc.vocabularies.impl.IPTCVocabularyImpl.EntryImpl;
import javax.annotation.Nonnull;

/*******************************************************************************
 *
 * @author  Fabrizio Giudici
 * @version $Id$
 *
 ******************************************************************************/
public class IPTCVocabularyBuilder 
  {
    private static final XPathFactory XPATH_FACTORY = XPathFactory.newInstance();
    private static final XPath XPATH = XPATH_FACTORY.newXPath();
    private static final XPathExpression TOPIC_EXPRESSION;
    private static final XPathExpression LANGUAGE_EXPRESSION;
    private static final XPathExpression VARIANT_EXPRESSION;
    private static final XPathExpression TEXT_EXPRESSION;
    
    static
      {
        try
          {
            TOPIC_EXPRESSION = XPATH.compile("//NewsML/NewsItem/TopicSet/Topic");
            LANGUAGE_EXPRESSION = XPATH.compile("@xml:lang");
            VARIANT_EXPRESSION = XPATH.compile("@Variant");
            TEXT_EXPRESSION = XPATH.compile(".");
          }
        catch (XPathExpressionException e)
          {
            throw new ExceptionInInitializerError(e);
          }
      }

    private static final EntityResolver entityResolver = new EntityResolver() 
      {
        public InputSource resolveEntity (@Nonnull final String publicId,
                                          @Nonnull final String systemId)
          throws SAXException, IOException 
          {
            if ("urn:newsml:iptc.org:20001006:NewsMLv1.0.dtd:1".equals(publicId))
              {
                final InputStream is = getClass().getResourceAsStream("datasource/NewsMLv1.0.dtd.gz");
                final GZIPInputStream gis = new GZIPInputStream(is);
                return new InputSource(gis);
              }
            
            return null;
          }
      };

    @Nonnull
    public static IPTCVocabularyImpl createVocabulary (@Nonnull final String resourceName)
      {
        try
          {
            final IPTCVocabularyImpl vocabulary = new IPTCVocabularyImpl();
            final InputStream is = vocabulary.getClass().getResourceAsStream("datasource/" + resourceName);
            final InputSource source = new InputSource(new GZIPInputStream(is));
            final DocumentBuilderFactory builderFactory = DocumentBuilderFactory.newInstance();
            final DocumentBuilder builder = builderFactory.newDocumentBuilder();
            builder.setEntityResolver(entityResolver);
            final Document document = builder.parse(source);
            is.close();

            final NodeList topicNodes = (NodeList)TOPIC_EXPRESSION.evaluate(document, XPathConstants.NODESET);
            final Map<String, String> descriptionMap = new HashMap<String, String>();

            for (int i = 0; i < topicNodes.getLength(); i++)
              {
                final Element topicElement = (Element)topicNodes.item(i);  
                final NodeList descriptionNodes = topicElement.getElementsByTagName("Description");
                final String formalName = null;
                descriptionMap.clear();

                for (int j = 0; j < descriptionNodes.getLength(); j++)
                  {
                    final Element descriptionElement = (Element)descriptionNodes.item(j);  
//                    final String language = (String)LANGUAGE_EXPRESSION.evaluate(descriptionElement, XPathConstants.STRING);
                    final String variant = (String)VARIANT_EXPRESSION.evaluate(descriptionElement, XPathConstants.STRING);
                    final String text = (String)TEXT_EXPRESSION.evaluate(descriptionElement, XPathConstants.STRING);

//                    System.err.println("L:" + language + " V:" + variant + " T:" + text);
                    descriptionMap.put(variant, text);
                  }

                final Entry entry = new EntryImpl(formalName, 
                                                  descriptionMap.get("Name"), 
                                                  descriptionMap.get("Explanation"), 
                                                  descriptionMap.get("ChangeComment"), 
                                                  descriptionMap.get("ChangeVersion"), 
                                                  descriptionMap.get("FirstVersion"), 
                                                  descriptionMap.get("DeprecatedVersion"));
                vocabulary.add(entry);
              }

            vocabulary.prepare();

            return vocabulary;
          }
        catch (RuntimeException e)
          {
            throw e;
          }
        catch (Exception e)
          {
            throw new RuntimeException(e);
          }
      }
    
    private static void generateVocabulary (@Nonnull final String inName,
                                            @Nonnull final String outName)
      throws IOException
      {
        final IPTCVocabularyImpl vocabulary = createVocabulary(inName + ".gz");
        final File file = new File(outName + ".gz");
        System.err.println("Generating " + file.getAbsolutePath() + "...");
        final FileOutputStream os = new FileOutputStream(file);
        final GZIPOutputStream gos = new GZIPOutputStream(os);
        final ObjectOutputStream oos = new ObjectOutputStream(gos);
        oos.writeObject(vocabulary.internalGetEntries());
        oos.close();
      }
    
    public static void main (final String ... args)
      throws IOException
      {
        final String base = args[0] + "/src/it/tidalwave/metadata/iptc/vocabularies/impl/datasource/";
        generateVocabulary("topicset.iptc-genre.xml", base + "topicset.iptc-genre.ser");
        generateVocabulary("topicset.iptc-provider.xml", base + "topicset.iptc-provider.ser");
        generateVocabulary("topicset.iptc-scene.xml", base + "topicset.iptc-scene.ser");
        generateVocabulary("topicset.iptc-subjectcode.xml", base + "topicset.iptc-subjectcode.ser");
      }
  }
